#!/usr/bin/python

## expendable

## Copyright (C) 2008, 2009 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import sys
if __name__ == '__main__':
    sys.path.append (os.environ.get ('top_srcdir', '.'))

from gettext import gettext as _
import gtk.glade
import locale
from numpy.lib import financial
import account

name=_("Mortgage")
description=_("Repayment mortgage with optional offset facility.")

class Model(account.Model):
    def __init__ (self, rate, loan_amount, nper, offset_amount=0):
        account.Model.__init__ (self)
        self.int_rate_apr = rate
        self.mrate = self.int_rate_apr / 12

        self.loan_amount = loan_amount
        self.offset_amount = offset_amount
        self.nper = nper

        self.paid = 0
        self.int_paid = 0

        self.pmt = financial.pmt (self.mrate,
                                  self.nper,
                                  self.loan_amount)
        self.pmt = int (100 * self.pmt - 0.9) * 0.01

        # Calculate a full mortgage term without offset for comparison
        # of interest paid.
        int_paid_no_offset = 0
        for per in xrange (nper):
            ipmt = -loan_amount * self.mrate
            ipmt = int (100 * ipmt) * 0.01
            ppmt = self.pmt - ipmt
            ppmt = int (100 * ppmt) * 0.01
            loan_amount += ppmt
            loan_amount = int (100 * loan_amount) * 0.01
            int_paid_no_offset -= ipmt
        self.int_paid_no_offset = int_paid_no_offset

    def cash_available (self):
        return account.Model.cash_available (self)

    def test_adjust (self, amount, by_account):
        return self.adjust (amount, by_account, test=True)

    def adjust (self, amount, by_account, test=False):
        if self.offset_amount + amount < 0:
            amount = -self.offset_amount

        if not test:
            self.offset_amount += amount

        return -amount

    def _record (self, ipmt=0, ppmt=0):
        this_data = { 'per': self.per,
                      'loan amount': self.loan_amount,
                      'offset': self.offset_amount,
                      'total interest': self.int_paid,
                      'total interest saved': (self.int_paid_no_offset -
                                               self.int_paid),
                      'total repayments': self.paid,
                      'pmt': self.pmt,
                      'ipmt': ipmt,
                      'ppmt': ppmt }
        self.model_data.append (this_data)

    def iter (self):
        if self.per == 0:
            # First month.
            self._record ()
            self.nper -= 1
            return False

        if self.loan_amount == 0:
            return True

        ipmt = -self.loan_amount * self.mrate
        ipmt = int (100 * ipmt) * 0.01
        ppmt = self.pmt - ipmt
        ppmt = int (100 * ppmt) * 0.01

        oipmt = self.offset_amount * self.mrate
        oipmt = int (100 * oipmt) * 0.01

        if self.loan_amount <= self.offset_amount:
            ipmt = ppmt = oipmt = 0

        if self.parent != None:
            self.parent.adjust (self.pmt, self)

        self.loan_amount += ppmt
        self.paid += -self.pmt
        self.int_paid -= ipmt + oipmt
        self.nper -= 1

        self.offset_amount += oipmt

        # If we have enough cash to cancel out the mortgage and cope
        # with our current liabilities, do it.
        cash_available = self.cash_available ()
        liabilities = 0
        if cash_available < 0:
            liabilities = -cash_available

        if self.offset_amount - liabilities >= self.loan_amount:
            self.offset_amount -= self.loan_amount
            self.loan_amount = 0

        # Perform rounding.
        self.offset_amount = int (100 * self.offset_amount) * 0.01
        self.loan_amount = int (100 * self.loan_amount) * 0.01

        if self.loan_amount < 0.01:
            self.loan_amount = 0

        # Tidy up effects of previous rounding in the final month.
        if self.nper == 0:
            self.pmt = self.offset_amount + ipmt - self.loan_amount

        self._record (-ipmt, -ppmt)

        # Work out if we can cancel out the mortgage next month.  If
        # so, call in all our assets.
        outstanding = self.loan_amount - self.offset_amount
        if outstanding < 0:
            outstanding = 0

        # Calculate liabilities for two months (the month we call in
        # assets, and the month we pay off the mortgage with them).
        liabilities = 2 * self.cash_available_next_month ()
        if cash_available + liabilities >= outstanding:
            for model in self.models:
                try:
                    adjustment = model.adjust (-outstanding, self)
                    self.adjust (adjustment, self)
                    outstanding -= adjustment
                except RuntimeError:
                    # Not implemented by this sub-account.
                    pass

                if outstanding < 0.01:
                    break

        return False

    def get_datasets (self):
        model_data = self.model_data
        mdata = map (lambda x: (x['per'], x['loan amount']), model_data)
        odata = map (lambda x: (x['per'], x['offset']), model_data)
        idata = map (lambda x: (x['per'], x['total interest']), model_data)
        n = self.account.get_name ()
        return [{'name': n,
                 'data': mdata},
                {'name': _("%s (offset)") % n,
                 'data': odata},
                {'name': _("%s (interest)") % n,
                 'data': idata}]

class Data(account.Data):
    required_tags = ['rate', 'loan_amount', 'nper']
    glade = "mortgage.glade"

    def __init__ (self):
        account.Data.__init__ (self)
        for dir in sys.path:
            f = os.path.join (dir, self.glade)
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        data = {}
        for name in ['rate']:
            widget = gld.get_widget (name)
            data[name] = widget.get_value () / 100.0

        for name in ['loan_amount']:
            widget = gld.get_widget (name)
            data[name] = widget.get_value ()

        term = [gld.get_widget ('term_years'),
                gld.get_widget ('term_months')]
        data['nper'] = int ((term[0].get_value () * 12) +
                            term[1].get_value ())

        self.data = data

    def read (self, xml):
        data = {}
        for child in xml.getchildren ():
            # Floats
            if child.tag in ['rate',
                             'loan_amount',
                             'offset_amount']:
                data[child.tag] = float (child.text)
            # Ints
            elif child.tag in ['nper']:
                data[child.tag] = int (child.text)
            else:
                print "Not understood: %s" % child.tag

        for required in self.required_tags:
            if not data.has_key (required):
                raise RuntimeError, "Required tags: %s" % self.required_tags

        self.data = data

    def write (self):
        xml = "<mortgage>\n"
        for var in self.required_tags:
            value = self.data[var]
            if type (value) == float:
                value = "%.4f" % value
            xml += "  <%s>%s</%s>\n" % (var, value, var)
        for optvar in ['offset_amount']:
            v = self.data.get (optvar, None)
            if v != None:
                if type (v) == float:
                    v = "%.4f" % v
                xml += "  <%s>%s</%s>\n" % (optvar, v, optvar)
        xml += "</mortgage>\n"
        return xml

    def ui (self, variables_vbox, results_vbox):
        for dir in sys.path:
            f = os.path.join (dir, self.glade)
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        w = gld.get_widget ('variables_table')
        w.reparent (variables_vbox)
        rate = gld.get_widget ('rate')
        rate.set_value (self.data['rate'] * 100.0)
        rate.connect ("value-changed", self.rate_changed)

        for var in ['loan_amount', 'offset_amount']:
            widget = gld.get_widget (var)
            widget.set_value (self.data.get (var, 0))
            widget.connect ("value-changed", self.variable_changed, var)

        term = [gld.get_widget ('term_years'),
                gld.get_widget ('term_months')]
        term[0].set_value (self.data['nper'] / 12)
        term[1].set_value (self.data['nper'] % 12)
        self.term_widgets = term
        for widget in term:
            widget.connect ("value-changed", self.term_changed)

        w = gld.get_widget ('results_hbox')
        w.reparent (results_vbox)
        self.result_widget = {}
        for name in ['repaid', 'tint', 'pmt', 'term', 'saved', 'tsaved']:
            self.result_widget[name] = gld.get_widget (name)

    def rate_changed (self, rate):
        self.data['rate'] = rate.get_value () / 100.0
        try:
            self.redisplay ()
        except:
            widget.set_value (self.last_known_good[name] * 100.0)

    def variable_changed (self, widget, name):
        value = widget.get_value ()
        if name == 'term':
            value = int (value)
        self.data[name] = value
        try:
            self.redisplay ()
        except:
            widget.set_value (self.last_known_good[name])

    def term_changed (self, widget):
        set = widget.get_data ('set')
        widget.set_data ('set', False)
        term_years, term_months = self.term_widgets
        months = term_months.get_value ()
        years = term_years.get_value ()
        if months == 0 and years == 0:
            months = 1
            term_months.set_data ('set', True)
            term_months.set_value (months)
            term_years.set_value (years)
            return
        if months < 0:
            if years == 0:
                months = 1
            else:
                months = 12 + months
                years -= 1

            term_months.set_data ('set', True)
            term_months.set_value (months)
            term_years.set_value (years)
            return
        if months > 11:
            months = months - 12
            years += 1
            term_months.set_data ('set', True)
            term_months.set_value (months)
            term_years.set_value (years)
            return

        term = int (term_years.get_value () * 12 + term_months.get_value ())
        self.data['nper'] = term
        try:
            self.redisplay ()
        except:
            if not set:
                term = self.last_known_good['nper']
                years = term / 12
                months = term % 12
                term_months.set_data ('set', True)
                term_years.set_data ('set', True)
                term_months.set_value (months)
                term_years.set_value (years)

    def update_results (self, modeldata):
        n = len (modeldata)
        pmt = self.result_widget['pmt']
        tint = self.result_widget['tint']
        repaid = self.result_widget['repaid']
        term = self.result_widget['term']
        saved = self.result_widget['saved']
        tsaved = self.result_widget['tsaved']
        if n > 0:
            last = modeldata[n - 1]
            def time_format (f):
                return "%d yrs %d mths" % (f / 12, f % 12)

            pmt.set_text (locale.currency (-modeldata[0]['pmt'], True, True))
            tint.set_text (locale.currency (last['total interest'], True, True))
            repaid.set_text (locale.currency (last['total repayments'],
                                              True, True))
            term.set_text (time_format (last['per']))
            saved.set_text (locale.currency (last['total interest saved'],
                                             True, True))
            tsaved.set_text (time_format (self.data['nper'] - last['per']))
        else:
            pmt.set_text ('')
            tint.set_text ('')
            repaid.set_text ('')
            term.set_text ('')
            saved.set_text ('')
            tsaved.set_text ('')

    def get_model (self):
        return Model (**self.data)

if __name__ == '__main__':
    from pprint import *
    from xml.etree.ElementTree import XML
    d = Data ()
    w = d.write ()
    xml = XML (w)
    d.read (xml)

    xml = XML (
        "<mortgage name='foo'>\n"
        "  <rate>0.06</rate>\n"
        "  <loan_amount>20000</loan_amount>\n"
        "  <nper>%s</nper>\n"
        "</mortgage>\n" % (12 * 20))
    d.read (xml)
    XML (d.write ())
    m = d.get_model ()
    m.run ()
    #pprint (m.get_model_data ())

    def verify_data (rate, loan_amount, nper):
        xml = XML (
            "<mortgage name='foo'>\n"
            "  <rate>%s</rate>\n"
            "  <loan_amount>%s</loan_amount>\n"
            "  <nper>%s</nper>\n"
            "</mortgage>\n" % (rate, loan_amount, nper))
        d = Data ()
        d.read (xml)
        m = d.get_model ()
        m.run ()
        data = m.get_model_data ()

        repaid = False
        for d in data:
            if d['per'] == nper:
                if abs (d['loan amount']) < 0.01:
                    repaid = True

                if abs (d['total interest saved']) >= 0.01:
                    raise RuntimeError, "Incorrect interest calculation"

        if not repaid:
            pprint (data)
            raise RuntimeError, "Loan not repaid"

    verify_data (0.054, 68634, 180)
    verify_data (0.054, 68635, 180)
